/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: VectorImpl.h,v 1.3 2003/02/08 23:14:01 opetzold Exp $
 */

#ifndef TVMET_VECTOR_IMPL_H
#define TVMET_VECTOR_IMPL_H

#include <iomanip>			// setw

#include <tvmet/Functional.h>

namespace tvmet {

/*
 * member operators for i/o
 */
template<class T, std::size_t Sz>
inline
std::ostream& Vector<T, Sz>::print_on(std::ostream& os) const
{
  std::streamsize w = os.width();

  os << std::setw(0) << "Vector<" << typeid(T).name() << ", " << Sz << ">[";
  for(std::size_t i = 0; i < (Sz - 1); ++i) {
    os << std::setw(w) << m_data[i] << ", ";
  }
  os << std::setw(w) << m_data[Sz - 1] << "]";

  return os;
}


/*
 * member operators with scalars, per se element wise
 */
#define TVMET_IMPLEMENT_MEMBER(OP)					\
template<class T, std::size_t Sz>					\
inline									\
Vector<T, Sz>& Vector<T, Sz>::operator OP (value_type rhs) {		\
  using namespace element_wise;						\
  typedef XprLiteral<value_type> 			expr_type;	\
  *this OP XprVector<expr_type, Sz>(expr_type(rhs));			\
  return *this;								\
}

TVMET_IMPLEMENT_MEMBER(+=)
TVMET_IMPLEMENT_MEMBER(-=)
TVMET_IMPLEMENT_MEMBER(*=)
TVMET_IMPLEMENT_MEMBER(/=)
#undef TVMET_IMPLEMENT_MEMBER


#define TVMET_IMPLEMENT_MEMBER(OP)					\
template<class T, std::size_t Sz>					\
inline									\
Vector<T, Sz>& Vector<T, Sz>::operator OP (std::size_t rhs) {		\
  using namespace element_wise;						\
  typedef XprLiteral<value_type> 			expr_type;	\
  *this OP XprVector<expr_type, Sz>(expr_type(rhs));			\
  return *this;								\
}

TVMET_IMPLEMENT_MEMBER(%=)
TVMET_IMPLEMENT_MEMBER(^=)
TVMET_IMPLEMENT_MEMBER(&=)
TVMET_IMPLEMENT_MEMBER(|=)
TVMET_IMPLEMENT_MEMBER(<<=)
TVMET_IMPLEMENT_MEMBER(>>=)
#undef TVMET_IMPLEMENT_MEMBER


/*
 * member functions (operators) with vectors, for use with +=,-= ... <<=
 */
template<class T1, std::size_t Sz>
template <class T2>
inline Vector<T1, Sz>&
Vector<T1, Sz>::operator=(const Vector<T2, Sz>& rhs) {
  *this = XprVector<typename Vector<T2, Sz>::ConstReference, Sz>(rhs.cref());
  return *this;
}

#define TVMET_IMPLEMENT_MEMBER(NAME)					\
template<class T1, std::size_t Sz>					\
template <class T2>					                \
inline Vector<T1, Sz>& 					                \
Vector<T1, Sz>::NAME (const Vector<T2, Sz>& rhs) { 			\
  this->NAME( XprVector<typename Vector<T2, Sz>::ConstReference, Sz>(rhs.cref()) ); \
  return *this;					                	\
}

TVMET_IMPLEMENT_MEMBER(AddUpd)
TVMET_IMPLEMENT_MEMBER(SubUpd)
TVMET_IMPLEMENT_MEMBER(MulUpd)
TVMET_IMPLEMENT_MEMBER(DivUpd)
TVMET_IMPLEMENT_MEMBER(ModUpd)
TVMET_IMPLEMENT_MEMBER(BitwiseXorUpd)
TVMET_IMPLEMENT_MEMBER(BitwiseAndUpd)
TVMET_IMPLEMENT_MEMBER(BitwiseOrUpd)
TVMET_IMPLEMENT_MEMBER(ShlUpd)
TVMET_IMPLEMENT_MEMBER(ShrUpd)
#undef TVMET_IMPLEMENT_MEMBER


/*
 * member functions (operators) with expressions, for use width +=,-= ... <<=
 */
template<class T, std::size_t Sz>
template <class E>
inline Vector<T, Sz>&
Vector<T, Sz>::operator=(XprVector<E, Sz> expr) {
  assign(expr, fcnl_Assign<value_type, typename E::value_type>());
  return *this;
}

#define TVMET_IMPLEMENT_MEMBER(NAME, FCNL)				\
template<class T, std::size_t Sz>					\
template <class E>                                                    	\
inline Vector<T, Sz>&                                                  	\
Vector<T, Sz>::NAME (XprVector<E, Sz> expr) {				\
  assign(expr, FCNL<value_type, typename E::value_type>());         	\
  return *this;                                                       	\
}

TVMET_IMPLEMENT_MEMBER(AddUpd, fcnl_AddUpd)
TVMET_IMPLEMENT_MEMBER(SubUpd, fcnl_SubUpd)
TVMET_IMPLEMENT_MEMBER(MulUpd, fcnl_MulUpd)
TVMET_IMPLEMENT_MEMBER(DivUpd, fcnl_DivUpd)
TVMET_IMPLEMENT_MEMBER(ModUpd, fcnl_ModUpd)
TVMET_IMPLEMENT_MEMBER(BitwiseXorUpd, fcnl_BitwiseXorUpd)
TVMET_IMPLEMENT_MEMBER(BitwiseAndUpd, fcnl_BitwiseAndUpd)
TVMET_IMPLEMENT_MEMBER(BitwiseOrUpd, fcnl_BitwiseOrUpd)
TVMET_IMPLEMENT_MEMBER(ShlUpd, fcnl_ShlUpd)
TVMET_IMPLEMENT_MEMBER(ShrUpd, fcnl_ShrUpd)
#undef TVMET_IMPLEMENT_MEMBER

} // namespace tvmet

#endif // TVMET_VECTOR_IMPL_H

// Local Variables:
// mode:C++
// End:
