/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: Matrix.h,v 1.3 2003/02/08 23:14:01 opetzold Exp $
 */

#ifndef TVMET_META_MATRIX_H
#define TVMET_META_MATRIX_H

namespace tvmet {

namespace meta {

/**
 * \class Matrix Matrix.h "tvmet/meta/Matrix.h"
 * \brief Meta %Matrix class using expression templates
 */
template<std::size_t Rows, std::size_t Cols,
	 std::size_t RowStride=0, std::size_t ColStride=0>
class Matrix
{
public: // types
  typedef Matrix<Rows, Cols, RowStride, ColStride>	this_type;

private:
  Matrix();
  Matrix(const this_type&);
  Matrix& operator=(const this_type&);

private:
  enum {
    doRows = (RowStride < Rows - 1) ? 1 : 0,	/**< recursive counter Rows. */
    doCols = (ColStride < Cols - 1) ? 1 : 0	/**< recursive counter Cols. */
  };

public:
  /** Assign Operator on columns on given row. */
  template<class Matrx, class Expr, class Operator>
  static inline
  void assign2(Matrx& mat, Expr expr, Operator op) {
    op.applyOn(mat(RowStride,ColStride), expr(RowStride,ColStride));
    Matrix<Rows * doCols, Cols * doCols,
           RowStride * doCols, (ColStride+1) * doCols>::assign2(mat, expr, op);
  }

  /** Assign Operator on row-wise. */
  template<class Matrx, class Expr, class Operator>
  static inline
  void assign(Matrx& mat, Expr expr, Operator op) {
    Matrix<Rows, Cols,
           RowStride, 0>::assign2(mat, expr, op);
    Matrix<Rows * doRows, Cols * doRows,
          (RowStride+1) * doRows, 0>::assign(mat, expr, op);
  }

  /** evaluate a given matrix expression, column wise */
  template<class E>
  static inline
  bool all_elements2(const E& e) {
    if(!e(RowStride, ColStride)) return false;
    return Matrix<Rows * doCols, Cols * doCols,
                  RowStride * doCols, (ColStride+1) * doCols>::all_elements2(e);
  }

  /** evaluate a given matrix expression, row wise */
  template<class E>
  static inline
  bool all_elements(const E& e) {
    if(!Matrix<Rows, Cols, RowStride, 0>::all_elements2(e) ) return false;
    return Matrix<Rows * doRows, Cols * doRows,
                 (RowStride+1) * doRows, 0>::all_elements(e);
  }

  /** evaluate a given matrix expression, column wise */
  template<class E>
  static inline
  bool any_elements2(const E& e) {
    if(e(RowStride, ColStride)) return true;
    return Matrix<Rows * doCols, Cols * doCols,
                  RowStride * doCols, (ColStride+1) * doCols>::any_elements2(e);
  }

  /** evaluate a given matrix expression, row wise */
  template<class E>
  static inline
  bool any_elements(const E& e) {
    if(Matrix<Rows, Cols, RowStride, 0>::any_elements2(e) ) return true;
    return Matrix<Rows * doRows, Cols * doRows,
                 (RowStride+1) * doRows, 0>::any_elements(e);
  }

  /** evaluate a given matrix expression, column wise */
  template<class E>
  static inline
  bool eval2(const E& e) {
    if(!e(RowStride, ColStride)) return false;
    return Matrix<Rows * doCols, Cols * doCols,
                  RowStride * doCols, (ColStride+1) * doCols>::eval2(e);
  }

  /** evaluate a given matrix expression, row wise */
  template<class E>
  static inline
  bool eval(const E& e) {
    if(!Matrix<Rows, Cols, RowStride, 0>::eval2(e) ) return false;
    return Matrix<Rows * doRows, Cols * doRows,
                 (RowStride+1) * doRows, 0>::eval(e);
  }

  /** trace a given matrix expression */
  template<class E>
  static inline
  typename E::value_type
  trace(const E& e) {
    return e(RowStride, ColStride)
      + Matrix<Rows * doCols, Cols * doCols,
              (RowStride+1) * doCols, (ColStride+1) * doCols>::trace(e);
  }

};

/**
 * \class Matrix<0, 0, 0, 0> Matrix.h "tvmet/meta/Matrix.h"
 * \brief Meta %Matrix Specialized for recursion
 */
template<>
class Matrix<0, 0, 0, 0>
{
public:
  template<class Matrx, class Expr, class Operator>
  static inline void assign2(Matrx&, Expr, Operator) { }

  template<class Matrx, class Expr, class Operator>
  static inline void assign(Matrx&, Expr, Operator) { }

  template<class E>
  static inline bool all_elements2(const E&) { return true; }

  template<class E>
  static inline bool all_elements(const E&) { return true; }

  template<class E>
  static inline bool any_elements2(const E&) { return false; }

  template<class E>
  static inline bool any_elements(const E&) { return false; }

  template<class E>
  static inline bool eval2(const E&) { return true; }

  template<class E>
  static inline bool eval(const E&) { return true; }

  template<class E>
  static inline XprNull trace(const E&) { return XprNull(); }
};

} // namespace meta

} // namespace tvmet

#endif /* TVMET_META_MATRIX_H */

// Local Variables:
// mode:C++
// End:
