/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: Vector.h,v 1.4 2003/05/25 09:30:04 opetzold Exp $
 */

#ifndef TVMET_META_VECTOR_H
#define TVMET_META_VECTOR_H

namespace tvmet {

/* forwards */
template<class T, std::size_t Sz> class Vector;

namespace meta {

/**
 * \class Vector Vector.h "tvmet/meta/Vector.h"
 * \brief Meta %Vector class using expression templates
 */
template<std::size_t Sz, std::size_t Stride=0>
class Vector
{
public: // types
  typedef Vector<Sz, Stride>				this_type;

private:
  Vector();
  Vector(const this_type&);
  Vector& operator=(const this_type&);

private:
  enum {
    doIt = (Stride < (Sz-1)) ? 1 : 0		/**< recursive counter */
  };

public:
  /** assign on a vector using an operator an expression. This can be an
      ordinary operator like += or an assignment operator= which does a copy. */
  template <class Vec, class Expr, class Operator>
  static inline
  void assign(Vec& vec, Expr expr, Operator op) {
    op.applyOn(vec[Stride], expr[Stride]);
    meta::Vector<Sz * doIt, (Stride+1) * doIt>::assign(vec, expr, op);
  }

  /** build the sum of the vector. */
  template<class E>
  static inline
  typename E::value_type
  sum(const E& rhs) {
    return rhs[Stride] + meta::Vector<Sz * doIt, (Stride+1) * doIt>::sum(rhs);
  }

  /** build the sum using abs on all elemenets of the vector. */
  template<class E>
  static inline
  typename E::value_type
  abs_sum(const E& rhs) {
    return std::abs(rhs[Stride])
      + meta::Vector<Sz * doIt, (Stride+1) * doIt>::abs_sum(rhs);
  }

  /** build the product of the vector. */
  template<class E>
  static inline
  typename NumericTraits<typename E::value_type>::sum_type
  product(const E& rhs) {
    return rhs[Stride] * meta::Vector<Sz * doIt, (Stride+1) * doIt>::product(rhs);
  }

  /** build the dot product of the vector. */
  template<class E1, class E2>
  static inline
  typename PromoteTraits<typename E1::value_type,
			 typename E2::value_type>::value_type
  dot(const E1& lhs, const E2& rhs) {
    return lhs[Stride] * rhs[Stride]
      + meta::Vector<Sz * doIt, (Stride+1) * doIt>::dot(lhs, rhs);
  }

  /** build the cross product of the vector.
   * \todo Implement vector outer product as ET and MT, returning a XprVector
   */
  template<class T1, class T2>
  static inline
  tvmet::Vector<typename PromoteTraits<T1,T2>::value_type, Sz>
  cross(const tvmet::Vector<T1, Sz>& lhs, const tvmet::Vector<T2, Sz>& rhs) {
    typedef typename PromoteTraits<T1, T2>::value_type		value_type;
    return tvmet::Vector<value_type, Sz>(lhs[1]*rhs[2] - rhs[1]*lhs[2],
					 rhs[0]*lhs[2] - lhs[0]*rhs[2],
					 lhs[0]*rhs[1] - rhs[0]*lhs[1]);
  }

  /** check for all elements */
  template<class E>
  static inline
  bool
  all_elements(const E& e) {
    if(!e[Stride]) return false;
    return meta::Vector<Sz * doIt, (Stride+1) * doIt>::all_elements(e);
  }

  /** check for any elements */
  template<class E>
  static inline
  bool
  any_elements(const E& e) {
    if(e[Stride]) return true;
    return meta::Vector<Sz * doIt, (Stride+1) * doIt>::any_elements(e);
  }

  /** evaluate a given vector expression */
  template<class E>
  static inline
  bool
  eval(const E& e) {
    if(!e[Stride]) return false;
    return meta::Vector<Sz * doIt, (Stride+1) * doIt>::eval(e);
  }
};


/**
 * \class Vector<0,0> Vector.h "tvmet/meta/Vector.h"
 * \brief Meta %Vector Specialized for recursion
 */
template<>
class Vector<0,0> {
  Vector();
  Vector& operator=(const Vector&);

public:
  template <class Vec, class Expr, class Operator>
  static inline void assign(Vec&, Expr, Operator) { }

  template<class E>
  static inline XprNull sum(const E&) { return XprNull(); }

  template<class E>
  static inline XprNull abs_sum(const E&) { return XprNull(); }

  template<class E>
  static inline XprNull product(const E&) { return XprNull(); }

  template<class E1, class E2>
  static inline XprNull dot(const E1&, const E2&) { return XprNull(); }

  template<class E>
  static inline bool all_elements(const E& e) { return true; }

  template<class E>
  static inline bool any_elements(const E& e) { return false; }

  template<class E>
  static inline bool eval(const E& e) { return true; }
};

} // namespace meta

} // namespace tvmet

#endif /* TVMET_META_VECTOR_H */

// Local Variables:
// mode:C++
// End:
