/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: BinOperator.h,v 1.1 2003/03/20 19:47:24 opetzold Exp $
 */

#ifndef TVMET_XPR_BINOPERATOR_H
#define TVMET_XPR_BINOPERATOR_H

#include <tvmet/tvmet.h>
#include <tvmet/TypePromotion.h>

namespace tvmet {

/**
 * \class XprBinOp BinOperator.h "tvmet/xpr/BinOperator.h"
 * \brief Binary operators working on two sub expressions.
 *
 * On acessing using the index operator[] the binary operation will be
 * evaluated at compile time.
 */
template<class BinOp, class E1, class E2>
class XprBinOp  : public TvmetBase< XprBinOp<BinOp, E1, E2> >
{
  XprBinOp();
  XprBinOp& operator=(const XprBinOp&);

public:
  typedef typename
  PromoteTraits<typename E1::value_type,
		typename E2::value_type>::value_type	value_type;
  typedef E1						expr1_type;
  typedef E2						expr2_type;

public:
  /** Constructor for two expressions. */
#ifdef TVMET_OPTIMIZE_XPR_CTOR_ARGS_BY_VALUE
  explicit XprBinOp(expr1_type lhs, expr2_type rhs)
    : m_expr1(lhs), m_expr2(rhs) { }
#else
  explicit XprBinOp(const expr1_type& lhs, const expr2_type& rhs)
    : m_expr1(lhs), m_expr2(rhs) { }
#endif

  /** Copy Constructor. Not explicit! */
#ifdef TVMET_OPTIMIZE_XPR_MANUAL_CCTOR
  XprBinOp(const XprBinOp<BinOp, expr1_type, expr2_type>& bop)
    : m_expr1(bop.m_expr1), m_expr2(bop.m_expr2) { }
#endif

  /** Index operator, evaluates the expression inside. */
  value_type operator[](std::size_t i) const {
    return BinOp::applyOn(m_expr1[i], m_expr2[i]);
  }

  /** index operator for arrays/matrices */
  value_type operator()(std::size_t i, std::size_t j) const {
    return BinOp::applyOn(m_expr1(i, j), m_expr2(i, j));
  }

public: // debugging Xpr parse tree
  void print_on(std::ostream& os, std::size_t l=0) const {
    os << IndentLevel(l++) << "XprBinOp<" << std::endl;
    BinOp::print_on(os, l);
    m_expr1.print_on(os, l);
    m_expr2.print_on(os, l);
    os << IndentLevel(--l) << ">," << std::endl;
  }

private:
  expr1_type 						m_expr1;
  expr2_type 						m_expr2;
};

} // namespace tvmet

#endif // TVMET_XPR_BINOPERATOR_H

// Local Variables:
// mode:C++
// End:
