/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: MMProductTransposed.h,v 1.1 2003/05/17 14:05:39 opetzold Exp $
 */

#ifndef TVMET_XPR_MMPRODUCT_TRANSPOSED_H
#define TVMET_XPR_MMPRODUCT_TRANSPOSED_H

#include <tvmet/tvmet.h>
#include <tvmet/TypePromotion.h>

namespace tvmet {

/* forwards */
namespace meta {

template<std::size_t Rows1, std::size_t Cols1,
	 std::size_t Cols2,
	 std::size_t RowStride1, std::size_t ColStride1,
	 std::size_t RowStride2, std::size_t ColStride2,
	 std::size_t K>
class gemm;

} // namespace meta


/**
 * \class XprMMProductTransposed MMProductTransposed.h "tvmet/xpr/MMProductTransposed.h"
 * \brief Expression for transpose(matrix-matrix product).
 *        Using formula:
 *        \f[
 *        (M1*M2)^T
 *        \f]
 * \note The Rows2 has to be  equal to Cols1.
 */
template<class T1, std::size_t Rows1, std::size_t Cols1,
	 class T2, std::size_t Cols2,
	 std::size_t RowStride1, std::size_t ColStride1,
	 std::size_t RowStride2, std::size_t ColStride2>
class XprMMProductTransposed
  : public TvmetBase< XprMMProductTransposed<T1, Rows1, Cols1, T2, Cols2, RowStride1, ColStride1, RowStride2, ColStride2> >
{
private:
  XprMMProductTransposed();
  XprMMProductTransposed& operator=(const XprMMProductTransposed&);

public:
  typedef typename PromoteTraits<T1, T2>::value_type	value_type;

public:
  /** Constructor. */
  explicit XprMMProductTransposed(const T1* lhs,const T2* rhs)
    : m_lhs(lhs), m_rhs(rhs) { }

  /** Copy Constructor. Not explicit! */
#ifdef TVMET_OPTIMIZE_XPR_MANUAL_CCTOR
  XprMMProductTransposed(const XprMMProductTransposed<T1, Rows1, Cols1,
	       T2, Cols2,
	       RowStride1, ColStride1,
	       RowStride2, ColStride2>& mmp)
    : m_lhs(mmp.m_lhs), m_rhs(mmp.m_rhs) { }
#endif

  /** index operator for arrays/matrices */
  value_type operator()(std::size_t i, std::size_t j) const {
    return meta::gemm<Rows1, Cols1,
                     Cols2,
                     RowStride1, ColStride1,
                     RowStride2, ColStride2, 0>::product(m_lhs, m_rhs, j, i);
  }

public: // debugging Xpr parse tree
  void print_on(std::ostream& os, std::size_t l=0) const {
    os << IndentLevel(l++) << "XprMMProductTransposed<" << std::endl;
    os << IndentLevel(l) << typeid(T1).name() << ", " << Rows1 << ", " << Cols1 << ", ";
    os << typeid(T2).name() << ", " << Cols2 << ", ";
    os << RowStride1 << ", " << ColStride1 << ", ";
    os << RowStride2 << ", " << ColStride2 << std::endl;
    os << IndentLevel(--l) << ">," << std::endl;
  }

private:
  const T1* _tvmet_restrict 				m_lhs;
  const T2* _tvmet_restrict 				m_rhs;
};


} // namespace tvmet

#endif // TVMET_XPR_MMPRODUCT_TRANSPOSED_H

// Local Variables:
// mode:C++
// End:
