/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: Matrix.h,v 1.1 2003/03/20 19:47:24 opetzold Exp $
 */

#ifndef TVMET_XPR_MATRIX_H
#define TVMET_XPR_MATRIX_H

namespace tvmet {

/**
 * \class XprMatrix Matrix.h "tvmet/xpr/Matrix.h"
 * \brief Represents the expression for vectors at any node in the parse tree.
 *
 * Specifically, XprMatrix is the class that wraps the expression, and the
 * expression itself is represented by the template parameter E. The
 * class XprMatrix is known as an anonymizing expression wrapper because
 * it can hold any subexpression of arbitrary complexity, allowing
 * clients to work with any expression by holding on to it via the
 * wrapper, without having to know the name of the type object that
 * actually implements the expression.
 * \note leave the Ctors non-explicit to allow implicit type conversation.
 */
template<class E, std::size_t Rows, std::size_t Cols>
class XprMatrix : public TvmetBase< XprMatrix<E, Rows, Cols> >
{
  XprMatrix();
  XprMatrix& operator=(const XprMatrix&);

public:
  typedef E						expr_type;
  typedef typename expr_type::value_type		value_type;

public:
  /** Constructor. */
#ifdef TVMET_OPTIMIZE_XPR_CTOR_ARGS_BY_VALUE
  explicit XprMatrix(expr_type expr) : m_expr(expr) { }
#else
  explicit XprMatrix(const expr_type& expr) : m_expr(expr) { }
#endif

  /** Copy Constructor. Not explicit! */
#ifdef TVMET_OPTIMIZE_XPR_MANUAL_CCTOR
  XprMatrix(const XprMatrix<expr_type, Rows, Cols>& expr) : m_expr(expr.m_expr) { }
#endif

  /** index operator for arrays/matrizes */
  value_type operator()(std::size_t i, std::size_t j) const { return m_expr(i, j); }

public: // debugging Xpr parse tree
  void print_on(std::ostream& os, std::size_t l=0) const {
    os << IndentLevel(l++) << "XprMatrix<" << std::endl;
    m_expr.print_on(os, l);
    os << IndentLevel(l) << Rows << ", " << Cols << std::endl
       << IndentLevel(--l) << ">"
       << ((l != 0) ? "," : "") << std::endl;
  }

private:
  expr_type 						m_expr;
};

} // namespace tvmet

#endif // TVMET_XPR_MATRIX_H

// Local Variables:
// mode:C++
// End:
