/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: MatrixFunctions.h,v 1.6 2003/05/24 08:08:13 opetzold Exp $
 */

#ifndef TVMET_XPR_MATRIX_FUNCTIONS_H
#define TVMET_XPR_MATRIX_FUNCTIONS_H

namespace tvmet {

/* forwards */
template<class T, std::size_t Rows, std::size_t Cols> class Matrix;
template<class T, std::size_t Sz> class Vector;


/************************************************************************
 * product(XprMatrix<>, Xpr... ) functions
 ************************************************************************/


/**
 * \fn product(XprMatrix<E1, Rows1, Cols1> lhs, XprMatrix<E2, Cols1, Cols2> rhs)
 * \brief Evaluate the product of two XprMatrix.
 * \ingroup _binary_function
 * \note This function is using tempories for pre-evaluation of the expressions.
 */
template<class E1, std::size_t Rows1, std::size_t Cols1,
	 class E2, std::size_t Cols2>
inline
XprMatrix<
  XprMMProduct<
    typename E1::value_type, Rows1, Cols1,
    typename E2::value_type, Cols2,
    Cols1, 1,
    Cols2, 1
  >,
  Rows1, Cols2
>
product(XprMatrix<E1, Rows1, Cols1> lhs, XprMatrix<E2, Cols1, Cols2> rhs) {
  typedef Matrix<typename E1::value_type, Rows1, Cols1>	temp_matrix1_type;
  typedef Matrix<typename E2::value_type, Cols2, Cols1>	temp_matrix2_type;

  typedef XprMMProduct<
    typename E1::value_type, Rows1, Cols1,
    typename E2::value_type, Cols2,
    Cols1, 1,
    Cols2, 1
  >							expr_type;
  return XprMatrix<expr_type, Rows1, Cols2>(expr_type(temp_matrix1_type(lhs).data(), temp_matrix2_type(rhs).data()));
}


/**
 * \fn product(XprMatrix<E1, Rows, Cols> lhs, XprVector<E2, Cols> rhs)
 * \brief Evaluate the product of XprMatrix and XprVector.
 * \ingroup _binary_function
 * \note This function is using tempories for pre-evaluation of the expressions.
 */
template<class E1, std::size_t Rows, std::size_t Cols,
	 class E2>
inline
XprVector<
  XprMVProduct<
    typename E1::value_type, Rows, Cols,
    Cols, 1,
    typename E2::value_type, 1
  >,
  Rows
>
product(XprMatrix<E1, Rows, Cols> lhs, XprVector<E2, Cols> rhs) {
  typedef Matrix<typename E1::value_type, Rows, Cols>	temp_matrix_type;
  typedef Vector<typename E2::value_type, Cols>		temp_vector_type;

  typedef XprMVProduct<
    typename E1::value_type, Rows, Cols,
    Cols, 1,
    typename E2::value_type, 1
  >							expr_type;
  return XprVector<expr_type, Rows>(expr_type(temp_matrix_type(lhs).data(), temp_vector_type(rhs).data()));
}


/************************************************************************
 * global binary math functions
 * documented by doc/functions_doc.cc
 ************************************************************************/


/*
 * XprMatrix<E1, Rows, Cols> and XprMatrix<E2, Rows, Cols>
 */
#define TVMET_BINARY_FUNCTION(FUNC, FUNCTIONAL)                 	\
template<class E1, std::size_t Rows, std::size_t Cols, class E2>	\
inline                 			      			      	\
XprMatrix<                    	      					\
  XprBinOp<                    			      			\
    FUNCTIONAL<typename E1::value_type, typename E2::value_type>,	\
    XprMatrix<E1, Rows, Cols>,         			      		\
    XprMatrix<E2, Rows, Cols>         			      		\
  >,                   			      			      	\
  Rows, Cols           			      			      	\
>                   			      		      		\
FUNC(XprMatrix<E1, Rows, Cols> lhs, XprMatrix<E2, Rows, Cols> rhs) {	\
  typedef XprBinOp<							\
    FUNCTIONAL<typename E1::value_type, typename E2::value_type>,	\
    XprMatrix<E1, Rows, Cols>,						\
    XprMatrix<E2, Rows, Cols>						\
  >		    					expr_type; 	\
  return XprMatrix<expr_type, Rows, Cols>(expr_type(lhs, rhs));		\
}

TVMET_BINARY_FUNCTION(atan2, fcnl_atan2)
TVMET_BINARY_FUNCTION(drem, fcnl_drem)
TVMET_BINARY_FUNCTION(fmod, fcnl_fmod)
TVMET_BINARY_FUNCTION(hypot, fcnl_hypot)
TVMET_BINARY_FUNCTION(jn, fcnl_jn)
TVMET_BINARY_FUNCTION(yn, fcnl_yn)
TVMET_BINARY_FUNCTION(pow, fcnl_pow)
#undef TVMET_BINARY_FUNCTION


/*
 * XprMatrix<E, Rows, Cols> and POD's
 */
#define TVMET_BINARY_FUNCTION(FUNC, FUNCTIONAL, TP)           		\
template<class E, std::size_t Rows, std::size_t Cols>			\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, TP >,      			\
    XprMatrix<E, Rows, Cols>,		          			\
    XprLiteral< TP >          						\
  >,          								\
  Rows, Cols   								\
>          								\
FUNC(XprMatrix<E, Rows, Cols> lhs, TP rhs) {				\
  typedef XprBinOp<          						\
    FUNCTIONAL<typename E::value_type, TP >,   				\
    XprMatrix<E, Rows, Cols>,		        			\
    XprLiteral< TP >              			      		\
  >							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(expr_type(lhs, XprLiteral< TP >(rhs))); \
}          								\
          								\
template<class E, std::size_t Rows, std::size_t Cols>			\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, TP >,      			\
    XprLiteral< TP >,          						\
    XprMatrix<E, Rows, Cols>		          			\
  >,          								\
  Rows, Cols   								\
>          								\
FUNC(TP lhs, XprMatrix<E, Rows, Cols> rhs) {				\
  typedef XprBinOp<          						\
    FUNCTIONAL<typename E::value_type, TP >,   				\
    XprLiteral< TP >,              			      		\
    XprMatrix<E, Rows, Cols>		        			\
  >							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(expr_type(XprLiteral< TP >(lhs), rhs)); \
}

TVMET_BINARY_FUNCTION(atan2, fcnl_atan2, int)
TVMET_BINARY_FUNCTION(drem, fcnl_drem, int)
TVMET_BINARY_FUNCTION(fmod, fcnl_fmod, int)
TVMET_BINARY_FUNCTION(hypot, fcnl_hypot, int)
TVMET_BINARY_FUNCTION(jn, fcnl_jn, int)
TVMET_BINARY_FUNCTION(yn, fcnl_yn, int)
TVMET_BINARY_FUNCTION(pow, fcnl_pow, int)

TVMET_BINARY_FUNCTION(atan2, fcnl_atan2, float)
TVMET_BINARY_FUNCTION(drem, fcnl_drem, float)
TVMET_BINARY_FUNCTION(fmod, fcnl_fmod, float)
TVMET_BINARY_FUNCTION(hypot, fcnl_hypot, float)
TVMET_BINARY_FUNCTION(jn, fcnl_jn, float)
TVMET_BINARY_FUNCTION(yn, fcnl_yn, float)
TVMET_BINARY_FUNCTION(pow, fcnl_pow, float)

TVMET_BINARY_FUNCTION(atan2, fcnl_atan2, double)
TVMET_BINARY_FUNCTION(drem, fcnl_drem, double)
TVMET_BINARY_FUNCTION(fmod, fcnl_fmod, double)
TVMET_BINARY_FUNCTION(hypot, fcnl_hypot, double)
TVMET_BINARY_FUNCTION(jn, fcnl_jn, double)
TVMET_BINARY_FUNCTION(yn, fcnl_yn, double)
TVMET_BINARY_FUNCTION(pow, fcnl_pow, double)

#ifdef TVMET_HAVE_LONG_DOUBLE
TVMET_BINARY_FUNCTION(atan2, fcnl_atan2, long double)
TVMET_BINARY_FUNCTION(drem, fcnl_drem, long double)
TVMET_BINARY_FUNCTION(fmod, fcnl_fmod, long double)
TVMET_BINARY_FUNCTION(hypot, fcnl_hypot, long double)
TVMET_BINARY_FUNCTION(jn, fcnl_jn, long double)
TVMET_BINARY_FUNCTION(yn, fcnl_yn, long double)
TVMET_BINARY_FUNCTION(pow, fcnl_pow, long double)
#endif // TVMET_HAVE_LONG_DOUBLE

#undef TVMET_BINARY_FUNCTION


#ifdef TVMET_HAVE_COMPLEX
/*
 * XprMatrix<E, Rows, Cols> and std::complex<>
 */
#define TVMET_BINARY_FUNCTION(FUNC, FUNCTIONAL)	           		\
template<class E, std::size_t Rows, std::size_t Cols, class T>		\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, std::complex<T> >,  		\
    XprMatrix<E, Rows, Cols>,		          			\
    XprLiteral< std::complex<T> >          				\
  >,          								\
  Rows, Cols   								\
>          								\
FUNC(XprMatrix<E, Rows, Cols> lhs, const std::complex<T>& rhs) {	\
  typedef XprBinOp<          						\
    FUNCTIONAL<typename E::value_type, std::complex<T> >, 		\
    XprMatrix<E, Rows, Cols>,		        			\
    XprLiteral< std::complex<T> >         		      		\
  >							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(expr_type(lhs, XprLiteral< std::complex<T> >(rhs))); \
}

TVMET_BINARY_FUNCTION(atan2, fcnl_atan2)
TVMET_BINARY_FUNCTION(drem, fcnl_drem)
TVMET_BINARY_FUNCTION(fmod, fcnl_fmod)
TVMET_BINARY_FUNCTION(hypot, fcnl_hypot)
TVMET_BINARY_FUNCTION(jn, fcnl_jn)
TVMET_BINARY_FUNCTION(yn, fcnl_yn)
TVMET_BINARY_FUNCTION(pow, fcnl_pow)
#undef TVMET_BINARY_FUNCTION
#endif // TVMET_HAVE_COMPLEX


/************************************************************************
 * global unary math functions
 * documented by doc/functions_doc.cc
 ************************************************************************/

/*
 * Functions on XprMatrix<E, Rows, Cols>
 */
#define TVMET_UNARY_FUNCTION(FUNC, FUNCTIONAL)				\
template<class E, std::size_t Rows, std::size_t Cols>			\
inline									\
XprMatrix<								\
  XprUnOp<								\
    FUNCTIONAL<typename E::value_type>, 				\
    XprMatrix<E, Rows, Cols>						\
  >, 									\
  Rows, Cols								\
>									\
FUNC(XprMatrix<E, Rows, Cols> rhs) {					\
  typedef XprUnOp<							\
    FUNCTIONAL<typename E::value_type>, 				\
    XprMatrix<E, Rows, Cols>						\
  > 							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(expr_type(rhs));		\
}

TVMET_UNARY_FUNCTION(abs, fcnl_abs)
TVMET_UNARY_FUNCTION(cbrt, fcnl_cbrt)
TVMET_UNARY_FUNCTION(floor, fcnl_floor)
TVMET_UNARY_FUNCTION(rint, fcnl_rint)
TVMET_UNARY_FUNCTION(sin, fcnl_sin)
TVMET_UNARY_FUNCTION(sinh, fcnl_sinh)
TVMET_UNARY_FUNCTION(cos, fcnl_cos)
TVMET_UNARY_FUNCTION(cosh, fcnl_cosh)
TVMET_UNARY_FUNCTION(asin, fcnl_asin)
TVMET_UNARY_FUNCTION(acos, fcnl_acos)
TVMET_UNARY_FUNCTION(atan, fcnl_atan)
TVMET_UNARY_FUNCTION(exp, fcnl_exp)
TVMET_UNARY_FUNCTION(log, fcnl_log)
TVMET_UNARY_FUNCTION(log10, fcnl_log10)
TVMET_UNARY_FUNCTION(sqrt, fcnl_sqrt)
#ifdef TVMET_HAVE_IEEE_MATH
TVMET_UNARY_FUNCTION(asinh, fcnl_asinh)
TVMET_UNARY_FUNCTION(acosh, fcnl_acosh)
TVMET_UNARY_FUNCTION(atanh, fcnl_atanh)
TVMET_UNARY_FUNCTION(expm1, fcnl_expm1)
TVMET_UNARY_FUNCTION(log1p, fcnl_log1p)
TVMET_UNARY_FUNCTION(erf, fcnl_erf)
TVMET_UNARY_FUNCTION(erfc, fcnl_erfc)
// XXX default return is int !
TVMET_UNARY_FUNCTION(isnan, fcnl_isnan)
TVMET_UNARY_FUNCTION(isinf, fcnl_isinf)
TVMET_UNARY_FUNCTION(finite, fcnl_finite)
TVMET_UNARY_FUNCTION(j0, fcnl_j0)
TVMET_UNARY_FUNCTION(j1, fcnl_j1)
TVMET_UNARY_FUNCTION(y0, fcnl_y0)
TVMET_UNARY_FUNCTION(y1, fcnl_y1)
TVMET_UNARY_FUNCTION(lgamma, fcnl_lgamma)
#endif
#undef TVMET_UNARY_FUNCTION


} // namespace tvmet

#endif // TVMET_XPR_MATRIX_FUNCTIONS_H

// Local Variables:
// mode:C++
// End:
