/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: MtMProduct.h,v 1.1 2003/05/11 18:03:27 opetzold Exp $
 */

#ifndef TVMET_XPR_MTMPRODUCT_H
#define TVMET_XPR_MTMPRODUCT_H

#include <tvmet/tvmet.h>
#include <tvmet/TypePromotion.h>


namespace tvmet {

/* forwards */
namespace meta {

template<std::size_t Rows1, std::size_t Cols1,
	 std::size_t Cols2,
	 std::size_t RowStride1, std::size_t ColStride1,
	 std::size_t RowStride2, std::size_t ColStride2,
	 std::size_t K>
class gemtm;

} // namespace meta

/**
 * \class XprMtMProduct MtMProduct.h "tvmet/xpr/MtMProduct.h"
 * \brief Expression for product(matrix)-matrix product.
 *        using formula
 *        \f[
 *        M_1^{T}\,M_2
 *        \f]
 * \note The number of cols of matrix 2 have to be equal to number of rows of
 *       matrix 1, since matrix 1 is transposed - the result is a (Cols1 x Cols2)
 *       matrix.
 */
template<class T1, std::size_t Rows1, std::size_t Cols1,
	 class T2, std::size_t Cols2,
	 std::size_t RowStride1, std::size_t ColStride1,
	 std::size_t RowStride2, std::size_t ColStride2>
class XprMtMProduct
  : public TvmetBase< XprMtMProduct<T1, Rows1, Cols1, T2, Cols2, RowStride1, ColStride1, RowStride2, ColStride2> >
{
private:
  XprMtMProduct();
  XprMtMProduct& operator=(const XprMtMProduct&);

public:
  typedef typename PromoteTraits<T1, T2>::value_type	value_type;

public:
  /** Constructor. */
  explicit XprMtMProduct(const T1* lhs,const T2* rhs)
    : m_lhs(lhs), m_rhs(rhs) { }

  /** Copy Constructor. Not explicit! */
#ifdef TVMET_OPTIMIZE_XPR_MANUAL_CCTOR
  XprMtMProduct(const XprMtMProduct<T1, Rows1, Cols1,
	       T2, Cols2,
	       RowStride1, ColStride1,
	       RowStride2, ColStride2>& mmp)
    : m_lhs(mmp.m_lhs), m_rhs(mmp.m_rhs) { }
#endif

  /** index operator for arrays/matrices */
  value_type operator()(std::size_t i, std::size_t j) const {
    return meta::gemtm<Rows1, Cols1,
      Cols2,
      RowStride1, ColStride1,
      RowStride2, ColStride2, 0>::product(m_lhs, m_rhs, i, j);
  }

public: // debugging Xpr parse tree
  void print_on(std::ostream& os, std::size_t l=0) const {
    os << IndentLevel(l++) << "XprMtMProduct<" << std::endl;
    os << IndentLevel(l) << typeid(T1).name() << ", " << Rows1 << ", " << Cols1 << ", ";
    os << typeid(T2).name() << ", " << Cols2 << ", ";
    os << RowStride1 << ", " << ColStride1 << ", ";
    os << RowStride2 << ", " << ColStride2 << std::endl;
    os << IndentLevel(--l) << ">," << std::endl;
  }

private:
  const T1* _tvmet_restrict 				m_lhs;
  const T2* _tvmet_restrict 				m_rhs;
};

} // namespace tvmet

#endif // TVMET_XPR_MTMPRODUCT_H

// Local Variables:
// mode:C++
// End:
