/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: Vector.h,v 1.1 2003/03/20 19:47:24 opetzold Exp $
 */

#ifndef TVMET_XPR_VECTOR_H
#define TVMET_XPR_VECTOR_H

namespace tvmet {

/**
 * \class XprVector Vector.h "tvmet/xpr/Vector.h"
 * \brief Represents the expression for vectors at any node in the parse tree.
 *
 * Specifically, XprVector is the class that wraps the expression, and the
 * expression itself is represented by the template parameter E. The
 * class XprVector is known as an anonymizing expression wrapper because
 * it can hold any subexpression of arbitrary complexity, allowing
 * clients to work with any expression by holding on to it via the
 * wrapper, without having to know the name of the type object that
 * actually implements the expression.
 * \note leave the Ctors non-explicit to allow implicit type conversation.
 *
 * \todo We have a const correctness problem!? It occoured on introducing the swap
 * function for vectors last. (The matrix hasn't a swap yet). Normally operator[]
 * for const and non const access should be enough. Unfortunally we go into
 * trouble for construction of vectors now. Affected is the Vector(value_type rhs) and
 * the CommaInitializer class. The Vector(value_tyoe) constructor problem is
 * easily to solve by writing a simple loop which doesn't call any fcnl_Assign.
 * The CommaInitializer problem seems to be more complicated, no solution yet.
 * The problem results from getting a reference non-const operator[] opposite
 * to the const operator[] which returns an object.
 * The solution (quick & dirty) is to use the operator() for non-const access.
 */
template<class E, std::size_t Sz>
class XprVector : public TvmetBase< XprVector<E, Sz> >
{
  XprVector();
  XprVector& operator=(const XprVector&);

public:
  typedef E						expr_type;
  typedef typename expr_type::value_type		value_type;

public:
  /** Constructor. */
#ifdef TVMET_OPTIMIZE_XPR_CTOR_ARGS_BY_VALUE
  explicit XprVector(expr_type expr) : m_expr(expr) { }
#else
  explicit XprVector(const expr_type& expr) : m_expr(expr) { }
#endif

  /** Copy Constructor. Not explicit! */
#ifdef TVMET_OPTIMIZE_XPR_MANUAL_CCTOR
  XprVector(const XprVector<expr_type, Sz>& expr) : m_expr(expr.m_expr) { }
#endif

  /** const index operator for vectors. */
  value_type operator[](std::size_t i) const {
    //std::cout << "XprVector::operator[] const\n";
    return m_expr[i];
  }

  /** index operator for vectors. */
  value_type& operator()(std::size_t i) {
    //std::cout << "XprVector::operator[]\n";
    return m_expr[i];
  }

public: // debugging Xpr parse tree
  void print_on(std::ostream& os, std::size_t l=0) const {
    os << IndentLevel(l++) << "XprVector<" << std::endl;
    m_expr.print_on(os, l);
    os << IndentLevel(l) << Sz << std::endl
       << IndentLevel(--l) << ">"
       << ((l != 0) ? "," : "") << std::endl;
  }

private:
  expr_type 						m_expr;
};

} // namespace tvmet

#endif // TVMET_XPR_VECTOR_H

// Local Variables:
// mode:C++
// End:
